/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "include/liblogger/register_service.hh"

#include "include/liblogger/exceptions.hh"

#include "src/util.hh"

#include "libpg/query.hh"
#include "libpg/sql_state/code.hh"

#include <boost/algorithm/string.hpp>

#include <regex>

namespace Fred {
namespace LibLogger {

namespace {

using PartitionPostfix = LibPg::Util::StrongType<std::string, struct PartitionPostfixTag_>;

PartitionPostfix make_partition_postfix_from_service_handle(const LogEntry::ServiceHandle& _service_handle)
{
    return PartitionPostfix{_service_handle.get() + "_"};
}

} // namespace Fred::LibLogger::{anonymous}

bool is_valid(const LogEntry::ServiceHandle& _service_handle)
{
    const std::regex service_handle_regex{"[a-z0-9_]{1,9}"};
    return std::regex_match(_service_handle.get(), service_handle_regex);
}

void register_service(
        const LibPg::PgRwTransaction& _rw_transaction,
        const LogEntry::ServiceName& _service_name,
        const LogEntry::ServiceHandle& _service_handle)
{
    if (!is_valid(_service_name) ||
        !is_valid(_service_handle))
    {
        throw InvalidService {};
    }

    const auto query = LibPg::make_query() <<
            // clang-format off
            "INSERT INTO service ("
               "partition_postfix, "
               "name) "
            "VALUES (" <<
                LibPg::parameter<PartitionPostfix>().as_text() << ", " <<
                LibPg::parameter<LogEntry::ServiceName>().as_text() << ") "
                "ON CONFLICT DO NOTHING";
            // clang-format on

    const auto service_partition_postfix = make_partition_postfix_from_service_handle(_service_handle);

    try
    {
        LibPg::PgResultNoData{
            exec(_rw_transaction, query, {service_partition_postfix, _service_name})};
    }
    catch (const LibPg::ExecFailure& e)
    {
        if (e.error_code() == LibPg::SqlState::string_data_right_truncation)
        {
            throw InvalidService{};
        }
        throw;
    }
}

} // namespace Fred::LibLogger
} // namespace Fred
